#include <stdio.h>
#include <stdlib.h>
#include <math.h>

/**********************************************************************************************************
 * Numerical integration code for a particle in a Z symmetrical magnetic field,
 * with applied electric field.  RK4, Euler, Backward, & Partitioned Euler
 * methods.
 * Thomas Hayward
 * TUM
 * November/December 2012
**********************************************************************************************************/


#define VPERP 1.0e7

struct vector3Struct {
	double a[3];
};
struct state3Struct {
	double r[3];
	double v[3];
};

typedef struct vector3Struct Vector3;
typedef struct state3Struct State3;

int hydrogen_ion(Vector3 (*acceleration)(State3));
State3 initial(void);
Vector3 Ion_acceleration_W_E(State3 state);
Vector3 Ion_acceleration_WO_E(State3 state);

int oscillator(Vector3 (*acceleration)(State3));
Vector3 osc_accel(State3 state);
State3 osc_initial(void);
double osc_energy(State3 state);

State3 Integrate(State3 state, const double t0, const double t1, const double dt, char e, Vector3 (*acceleration)(State3));
State3 EulerE_evolve(State3 state, const double dt, Vector3 (*acceleration)(State3));
State3 EulerI_evolve(State3 state, const double dt, Vector3 (*acceleration)(State3));
State3 EulerS_evolve(State3 state, const double dt, Vector3 (*acceleration)(State3));
State3 RK4_evolve(State3 state, const double dt, Vector3 (*acceleration)(State3));
State3 RK4_k(State3 state, const State3 k, const double dt, const double kfrac, const double hfrac, Vector3 (*acceleration)(State3));
State3 stateSum(State3 a, State3 b, double afrac, double bfrac);

State3 Integrate(State3 state, const double t0, const double t1, const double dt, char e, Vector3 (*acceleration)(State3)) {
	double t=t0;
	long count = 0;
	while(t1-t>0.5*dt) {
		switch (e) {
			case 'i':
				state = EulerI_evolve(state, dt, acceleration);
				break;
			case 'e':
				state = EulerE_evolve(state, dt, acceleration);
				break;
			case 's':
				state = EulerS_evolve(state, dt, acceleration);
				break;
			case 'r':
				state = RK4_evolve(state, dt, acceleration);
		}
		t+=dt;
		if(count%100 == 0) {
			fprintf(stdout, "%1.3f %1.3f %1.3f %1.3f %1.3f %1.3f %1.3f %1.3f I:%c p67\n", t, state.r[0], state.r[1], state.r[2], state.v[0], state.v[1], state.v[2], 0.5*(pow(state.v[0],2)+pow(state.v[1],2)+pow(state.v[2],2)), e);
		}
		count++;
	}
	return state;
}

State3 EulerE_evolve(State3 state, const double dt, Vector3 (*acceleration)(State3))  {
	Vector3 acc;
	int i;
	acc = (*acceleration)(state);
	for (i=0; i<3; i++) {
		state.r[i] += state.v[i]*dt;
		state.v[i] += acc.a[i]*dt;
	}
	return state;
}

State3 EulerI_evolve(State3 state, const double dt, Vector3 (*acceleration)(State3)) {
	Vector3 acc, acc2;
	State3 temp;
	int i;
	acc = (*acceleration)(state);
	for (i=0; i<3; i++) {
		temp.v[i] = acc.a[i]*dt + state.v[i];
		state.r[i] += temp.v[i]*dt;
	}
	// temp only new v;
	acc2 = (*acceleration)(temp);
	for (i=0; i<3; i++) {
		state.v[i] += acc2.a[i]*dt;
	}
	return state;
}

State3 EulerS_evolve(State3 state, const double dt, Vector3 (*acceleration)(State3)) {
	Vector3 acc;
	int i;
	acc = (*acceleration)(state);
	for (i=0; i<3; i++) {
		state.v[i] += acc.a[i]*dt;
		state.r[i] += state.v[i]*dt;
	}
	return state;
}

State3 RK4_k(State3 state, const State3 kPrev, const double dt, const double kfrac, const double hfrac, Vector3 (*acceleration)(State3)) {
	Vector3 acc;
	State3 kNew;
	int i;
	for (i=0; i<3; i++) {
		state.r[i] += kPrev.r[i]*kfrac;
		state.v[i] += kPrev.v[i]*kfrac;
	}
	acc = (*acceleration)(state);

	for (i=0; i<3; i++) {
		kNew.r[i] = dt*state.v[i];
		kNew.v[i] = dt*acc.a[i];
	}
	//fprintf(stderr, "a%1.3f %1.3f %1.3f \t", kNew.v[0], kNew.v[1], kNew.v[2]);
	return kNew;
}

Vector3 Ion_acceleration_W_E(State3 state) {
	Vector3 f;
	Vector3 B;
	Vector3 E;
	// scaled E = E/v_perp
	E.a[0] = 1.0e5 / VPERP;
	E.a[1] = 0.0;
	E.a[2] = 0.0;

	B.a[0] = 0.0;
	B.a[1] = 0.0;
	B.a[2] = 1.0;

	f.a[0] = state.v[1] + E.a[0];
	f.a[1] = -state.v[0];
	f.a[2] = 0.0;
	return f;
}

Vector3 Ion_acceleration_WO_E(State3 state) {
	Vector3 f;
	f.a[0] = state.v[1];
	f.a[1] = -state.v[0];
	f.a[2] = 0.0;
	return f;
}

State3 stateSum(State3 a, State3 b, double afrac, double bfrac) {
	State3 out;
	int i;
	for (i=0; i<3; i++) {
		out.r[i] = afrac*a.r[i] + bfrac*b.r[i];
		out.v[i] = afrac*a.v[i] + bfrac*b.v[i];
	}
	return out;
}

State3 RK4_evolve(State3 state, const double dt, Vector3 (*acceleration)(State3)) {
	State3 k1 = RK4_k(state, state, dt, 0.0, 0.0, acceleration);
	State3 k2 = RK4_k(state, k1, dt, 0.5, 0.5, acceleration);
	State3 k3 = RK4_k(state, k2, dt, 0.5, 0.5, acceleration);
	State3 k4 = RK4_k(state, k3, dt, 1.0, 1.0, acceleration);
	const double sixth = (double)1.0/(double)6.0;

	return stateSum(state, stateSum(stateSum(k1, k4, 1.0, 1.0), stateSum(k2, k3, 2.0, 2.0), sixth, sixth), 1.0, 1.0);
}

State3 initial(void) {
	State3 state;
	state.r[0] = 0.0;
	state.r[1] = 1.0;
	state.r[2] = 0.0;
	state.v[0] = 1.0;
	state.v[1] = 0.0;
	state.v[2] = 1.0;
	return state;
}

int hydrogen_ion(Vector3 (*acceleration)(State3)) {
	State3 state;
	const double t0 = 0.0;
	double t1 = 2.0*M_PI*100.0;
	long N = 10;
	double dt;

	dt = 1.0/(double)N;

	state	= initial();
	state = Integrate(state, t0, t1, dt, 'r', acceleration);
	fprintf(stderr, "RK: time: %4.2f, dt: %1.4f, v final: %1.6f\n", t1, dt, sqrt(pow(state.v[0],2)+pow(state.v[1],2)));
	fprintf(stdout, "\n");

	state	= initial();
	state = Integrate(state, t0, t1, dt, 'e', acceleration);
	fprintf(stderr, "EE: time: %4.2f, dt: %1.4f, v final: %1.6f\n", t1, dt, sqrt(pow(state.v[0],2)+pow(state.v[1],2)));
	fprintf(stdout, "\n");

	state	= initial();
	state = Integrate(state, t0, t1, dt, 'i', acceleration);
	fprintf(stderr, "IE time: %4.2f, dt: %1.4f, v final: %1.6f\n", t1, dt, sqrt(pow(state.v[0],2)+pow(state.v[1],2)));

	state	= initial();
	state = Integrate(state, t0, t1, dt, 's', acceleration);
	fprintf(stderr, "SE time: %4.2f, dt: %1.4f, v final: %1.6f\n", t1, dt, sqrt(pow(state.v[0],2)+pow(state.v[1],2)));
	return EXIT_SUCCESS;
}

/*
 * part d
 */

State3 osc_initial(void) {
	State3 state;
	int i;
	state.r[0] = -1.;
	state.v[0] = 0.;


	for(i=1;i<3;i++) {
		state.r[i] = 0;
		state.v[i] = 0;
	}
	return state;
}

double osc_energy(State3 state) {
	// k_x = k_y = k_z = 1
	// m = 1
	double E = 0.;
	int i;
	for (i=0; i<1; i++) {
		E += pow(state.v[i],2);
		E += pow(state.r[i],2);
	}
	return 0.5*E;
}

int oscillator(Vector3 (*osc_accel)(State3)) {
	const double t0 = 0.0;
	double t1 = 20.0;
	long N = 100;
	double dt;
	State3 state;

	dt = 1.0/(double)N;

	state = osc_initial();
	fprintf(stderr, "EE: starting energy: %3f.\t", osc_energy(state));
	state = Integrate(state, t0, t1, dt, 'e', osc_accel);
	fprintf(stderr, "EE: time: %4.2f, dt: %1.4f, x final: %1.6f,  v final: %1.6f, Final Energy: %1.6f\n", t1, dt, state.r[0], state.v[0], osc_energy(state));
	fprintf(stdout, "\n");

	state = osc_initial();
	fprintf(stderr, "IE: starting energy: %3f.\t", osc_energy(state));
	state = Integrate(state, t0, t1, dt, 'i', osc_accel);
	fprintf(stderr, "IE: time: %4.2f, dt: %1.4f, x final: %1.6f,  v final: %1.6f, Final Energy: %1.6f\n", t1, dt, state.r[0], state.v[0], osc_energy(state));
	fprintf(stdout, "\n");

	state = osc_initial();
	fprintf(stderr, "SE: starting energy: %3f.\t", osc_energy(state));
	state = Integrate(state, t0, t1, dt, 's', osc_accel);
	fprintf(stderr, "SE: time: %4.2f, dt: %1.4f, x final: %1.6f,  v final: %1.6f, Final Energy: %1.6f\n", t1, dt, state.r[0], state.v[0], osc_energy(state));
	fprintf(stdout, "\n");

	state = osc_initial();
	fprintf(stderr, "RK: starting energy: %3f.\t", osc_energy(state));
	state = Integrate(state, t0, t1, dt, 'r', osc_accel);
	fprintf(stderr, "RK: time: %4.2f, dt: %1.4f, x final: %1.6f,  v final: %1.6f, Final Energy: %1.6f\n", t1, dt, state.r[0], state.v[0], osc_energy(state));

	return EXIT_SUCCESS;
}

Vector3 osc_accel(State3 state) {
	Vector3 acc;
	acc.a[0] = -state.r[0];
	return acc;
}

int main(int argc, char *argv[]) {
	hydrogen_ion(Ion_acceleration_WO_E);
	hydrogen_ion(Ion_acceleration_W_E);
	oscillator(osc_accel);
	return EXIT_SUCCESS;
}
